"""Classes to capture the WeatherXM Data Model."""

import datetime
from dataclasses import dataclass
from enum import StrEnum
from typing import Any


class DeviceRelation(StrEnum):
    """Whether the WeatherXM device is owned by the account or just followed."""

    OWNED = "owned"
    FOLLOWED = "followed"


class BatteryState(StrEnum):
    """The battery state of a WeatherXM device."""

    OK = "ok"
    LOW = "low"


@dataclass(frozen=True, kw_only=True)
class Location:
    """Represents a location on the Earth.

    Latitude and Longitude values are represented in decimal degrees.
    """

    latitude: float
    longitude: float

    @classmethod
    def unmarshal(cls, data: dict[str, float]) -> "Location":  # noqa: D102
        return Location(latitude=data["lat"], longitude=data["lon"])


@dataclass(kw_only=True)
class HourlyWeatherData:
    """Detailed weather data.

    Data is valid at the specified timestamp.
    """

    timestamp: datetime.datetime
    temperature: float
    """Temperature in degrees Celsius."""
    apparent_temperature: float
    """Apparent temperature in degrees Celsius."""
    dew_point: float
    """Dew point in degrees Celsius."""
    humidity: int
    """Humidity in %"""
    precipitation_rate: float
    """Current precipitation rate in mm/h."""
    precipitation_accumulated: float
    """Accumlated precipitation for the day in mm."""
    wind_speed: float
    """Wind speed in m/s."""
    wind_gust: float
    """Wind gust speed in m/s."""
    wind_direction: int
    """Originating wind direction in degrees."""
    absolute_pressure: float
    """Absolute Air Pressure in hPa."""
    uv_index: int
    solar_irradiance: float
    """Solar Irradiance in W/m2."""
    icon: str
    """String identifier for the current weather state.

    Possible icon values are not documented.
    """

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "HourlyWeatherData":  # noqa: D102
        return HourlyWeatherData(
            timestamp=datetime.datetime.fromisoformat(data["timestamp"]),
            temperature=data["temperature"],
            apparent_temperature=data["feels_like"],
            dew_point=data["dew_point"],
            humidity=data["humidity"],
            precipitation_rate=data["precipitation"],
            precipitation_accumulated=data["precipitation_accumulated"],
            wind_speed=data["wind_speed"],
            wind_gust=data["wind_gust"],
            wind_direction=data["wind_direction"],
            absolute_pressure=data["pressure"],
            uv_index=data["uv_index"],
            solar_irradiance=data["solar_irradiance"],
            icon=data["icon"],
        )


@dataclass(kw_only=True)
class WxmDevice:
    """Represents a WeatherXM Device."""

    id: str
    """Unique identifier for the device."""
    name: str
    """Public name for the device."""
    friendly_name: str | None
    """Optional friendly name for the device."""

    relation: DeviceRelation
    """Whether the device is owned by the account, or just followed."""

    weather_station_model: str
    """The model of the weather station."""
    firmware_version: str
    """Current firmware version of the weather station."""

    address: str
    """Address where the weather station is installed."""
    timezone: str
    """Timezone where the weather station is installed."""
    location: Location
    """Location of the weather station."""

    battery_state: BatteryState
    """Current battery state of the weather station."""

    current_weather: HourlyWeatherData
    """Latest weather data reported by the station."""

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "WxmDevice":  # noqa: D102
        attributes = data["attributes"]
        bundle = data["bundle"]
        return WxmDevice(
            id=data["id"],
            name=data["name"],
            friendly_name=attributes.get("friendlyName", None),
            relation=DeviceRelation(data["relation"]),
            weather_station_model=bundle["ws_model"],
            firmware_version=attributes["firmware"]["current"],
            address=data["address"],
            timezone=data["timezone"],
            location=Location.unmarshal(data["location"]),
            battery_state=BatteryState(data["bat_state"]),
            current_weather=HourlyWeatherData.unmarshal(data["current_weather"]),
        )


@dataclass(kw_only=True)
class HourlyForecast:
    """Weather forecast data for an hourly forecast."""

    timestamp: datetime.datetime
    temperature: float
    feels_like_temperature: float
    humidity: int
    pressure: float
    precipitation: float
    precipitation_probability: int
    wind_speed: float
    wind_direction: int
    uv_index: int
    icon: str

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "HourlyForecast":  # noqa: D102
        return HourlyForecast(
            timestamp=datetime.datetime.fromisoformat(data["timestamp"]),
            temperature=data["temperature"],
            feels_like_temperature=data["feels_like"],
            humidity=data["humidity"],
            pressure=data["pressure"],
            precipitation=data["precipitation"],
            precipitation_probability=data["precipitation_probability"],
            wind_speed=data["wind_speed"],
            wind_direction=data["wind_direction"],
            uv_index=data["uv_index"],
            icon=data["icon"],
        )


@dataclass(kw_only=True)
class DailyForecast:
    """Weather forecast data for a daily forecast."""

    forecast_date: datetime.date
    temperature_min: float
    temperature_max: float
    humidity: int
    pressure: float
    precipitation_probability: int
    precipitation_intensity: float
    precipitation_type: str
    wind_speed: float
    wind_direction: int
    uv_index: int
    icon: str

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "DailyForecast":  # noqa: D102
        return DailyForecast(
            forecast_date=datetime.datetime.fromisoformat(data["timestamp"]).date(),
            temperature_min=data["temperature_min"],
            temperature_max=data["temperature_max"],
            humidity=data["humidity"],
            pressure=data["pressure"],
            precipitation_probability=data["precipitation_probability"],
            precipitation_intensity=data["precipitation_intensity"],
            precipitation_type=data["precipitation_type"],
            wind_speed=data["wind_speed"],
            wind_direction=data["wind_direction"],
            uv_index=data["uv_index"],
            icon=data["icon"],
        )


@dataclass(kw_only=True)
class ForecastForDate:
    """Forecast data for a specific date.

    Depending on the request used, the forecast may contain hourly forecasts,
    daily forecasts, or both.
    """

    forecast_date: datetime.date
    timezone: str
    hourly_forecasts: list[HourlyForecast] | None
    daily_forecast: DailyForecast | None

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "ForecastForDate":  # noqa: D102
        hourly_data: list[dict[str, Any]] = data.get("hourly", [])
        hourly_forecasts = [HourlyForecast.unmarshal(d) for d in hourly_data]

        daily_data = data.get("daily")
        daily_forecast = None
        if daily_data:
            daily_forecast = DailyForecast.unmarshal(daily_data)

        return ForecastForDate(
            forecast_date=datetime.date.fromisoformat(data["date"]),
            timezone=data["tz"],
            hourly_forecasts=hourly_forecasts,
            daily_forecast=daily_forecast,
        )


@dataclass
class WeatherForecast:
    """Encapsulates a WeatherXM weather forecast."""

    forecast: list[ForecastForDate]

    @classmethod
    def unmarshal(cls, data: list[dict[str, Any]]) -> "WeatherForecast":  # noqa: D102
        return WeatherForecast(
            forecast=[ForecastForDate.unmarshal(d) for d in data],
        )


@dataclass
class TokenSummary:
    """Token summary for a reward event."""

    timestamp: datetime.datetime
    base_reward: float
    total_business_boost_reward: float | None
    total_reward: float
    """Sum of base reward plus any boosts."""
    base_reward_score: int
    """Percentage of maximum base reward awarded to the station."""

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "TokenSummary":  # noqa: D102
        return TokenSummary(
            timestamp=datetime.datetime.fromisoformat(data["timestamp"]),
            base_reward=data["base_reward"],
            total_business_boost_reward=data.get("total_business_boost_reward"),
            total_reward=data["total_reward"],
            base_reward_score=data["base_reward_score"],
        )


@dataclass
class DeviceRewards:
    """Reward information for a device."""

    total_rewards: float
    latest_reward: TokenSummary

    @classmethod
    def unmarshal(cls, data: dict[str, Any]) -> "DeviceRewards":  # noqa: D102
        return DeviceRewards(
            total_rewards=data["total_rewards"],
            latest_reward=TokenSummary.unmarshal(data["latest"]),
        )



"""Classes wrapping the WeatherXM API."""

import asyncio
import datetime
import logging
from collections.abc import Awaitable, Callable
from enum import Enum, auto
from typing import Any

import aiohttp
import jwt
from yarl import URL

from .model import DeviceRewards, WeatherForecast, WxmDevice

_BASE_URL = URL("https://api.weatherxm.com/api/v1")

_LOGGER = logging.getLogger(__name__)


class WxmError(Exception):
    """Base exception class for errors from the WeatherXM API."""

    def __init__(self, message: str) -> None:
        """Initialise the WeatherXM error."""
        super().__init__(message)
        self.message = message


class AuthenticationError(WxmError):
    """Raised when an authentication error occurs.

    Authentication errors may occur during initial login, or if a problem occurs
    with token based authentication during other API interactions.
    """


class UnexpectedError(WxmError):
    """Raised when the API returns an unexpected error response.

    In most cases, simply retrying the query after a delay is the most
    appropriate response.
    """


RefreshTokenSubscriber = Callable[[str], Awaitable[Any]]
"""An interface for subscribing to notifications when the Refresh Token is updated.

The subscriber will be passed the updated Refresh Token."""


class WxmClient:
    """The WeatherXM Client.

    The Client should be used in conjunction with the WxmApi to query the WeatherXM API.
    """

    def __init__(
        self,
        session: aiohttp.ClientSession,
        refresh_token: str | None = None,
    ) -> None:
        """Initialises the WxmClient.

        Args:
            session: An aiohttp ClientSession instance. The session can be
                shared with other APIs.
        """
        self._session = session
        self.refresh_token = refresh_token

        self._access_token: str | None = None
        self._access_token_expiry: datetime.datetime | None = None
        self._token_subscribers: set[RefreshTokenSubscriber] = set()

    async def login(self, username: str, password: str) -> str:
        """Login into the WeatherXM API and generate a new access token.

        Logging in is only required if there is no valid refresh token.

        Returns:
            The new refresh token.
        """
        data = {
            "username": username,
            "password": password,
        }
        async with self._session.post(_BASE_URL / "auth/login", json=data) as resp:
            if resp.ok:
                access_tokens: dict[str, str] = await resp.json()
                await self._update_tokens(access_tokens)
                return access_tokens["refreshToken"]

            if resp.status in (400, 500):
                error = await resp.json()
                raise AuthenticationError(error["message"])
            raise AuthenticationError("Unknown error")

    async def get(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:  # noqa: ANN401
        """Send a get request to the WeatherXM API."""
        await self._refresh_access_token()

        headers = kwargs.pop("headers", {})
        headers["Authorization"] = f"Bearer {self._access_token}"
        return await self._session.get(_BASE_URL / url, headers=headers, **kwargs)

    async def subscribe_refresh_token(self, subscriber: RefreshTokenSubscriber) -> None:
        """Subscribe to receive notifications whenever the refresh token changes.

        Has no effect if the subscriber is already subscribed.
        """
        self._token_subscribers.add(subscriber)

    async def unsubscribe_refresh_token(
        self, subscriber: RefreshTokenSubscriber
    ) -> None:
        """Unsubscribe from receiving refresh token notifications.

        Has no effect if the subscriber is not subscribed.
        """
        self._token_subscribers.discard(subscriber)

    async def _refresh_access_token(self) -> None:
        if (
            self._access_token is not None
            and self._access_token_expiry is not None
            and self._access_token_expiry > datetime.datetime.now(tz=datetime.UTC)
        ):
            return

        if self._access_token:
            _LOGGER.debug("Access token expired, refreshing")

        data = {"refreshToken": self.refresh_token}
        async with self._session.post(_BASE_URL / "auth/refresh", json=data) as resp:
            match resp.status:
                case _ if resp.ok:
                    access_tokens: dict[str, str] = await resp.json()
                    await self._update_tokens(access_tokens)
                case 400:
                    error = await resp.json()
                    raise AuthenticationError(error["message"])
                # 5xx errors from the Cloudflare proxy don't indicate an
                # authentication error. Clients can retry later.
                case _ if resp.status >= 500 and resp.status < 600:  # noqa: PLR2004
                    _LOGGER.info("%d status when refreshing access token", resp.status)
                    _LOGGER.debug(
                        "Unexpected status '%d' during token refresh. Response: %s",
                        resp.status,
                        await resp.text(),
                    )
                    raise UnexpectedError(f"{resp.status} status returned")
                case _:
                    _LOGGER.error(
                        ("Unexpected status '%d' during token refresh. Response: %s"),
                        resp.status,
                        await resp.text(),
                    )
                    raise AuthenticationError(f"Unknown error: {resp.status}")

    async def _update_tokens(self, access_tokens: dict[str, str]) -> None:
        old_refresh_token = self.refresh_token

        self.refresh_token = access_tokens["refreshToken"]
        self._access_token = access_tokens["token"]

        # Decode the token expiry time
        payload = jwt.decode(
            self._access_token, options={"verify_signature": False, "require": ["exp"]}
        )
        self._access_token_expiry = datetime.datetime.fromtimestamp(
            payload["exp"], tz=datetime.UTC
        )
        _LOGGER.debug("Updated access token. New expiry: %s", self._access_token_expiry)

        # Notify subscribers if the refresh token has changed.
        if old_refresh_token and old_refresh_token != self.refresh_token:
            for coro in asyncio.as_completed(
                [s(self.refresh_token) for s in self._token_subscribers]
            ):
                try:
                    _ = await coro
                except Exception:
                    _LOGGER.exception("Exception from subscriber")


class ForecastType(Enum):
    """The type of weather forecast data to retreive."""

    HOURLY = auto()
    DAILY = auto()
    BOTH = auto()


class WxmApi:
    """API for querying specific data from the WeatherXM API."""

    def __init__(self, client: WxmClient) -> None:
        """Initialises the WeatherXM API.

        Args:
            client: An authenticated WxmClient.
        """
        self.client = client

    async def list_devices(self) -> list[WxmDevice]:
        """Get the list of WeatherXM devices associated with the current user."""
        async with await self.client.get("me/devices") as resp:
            await self._raise_if_error(resp)

            if _LOGGER.isEnabledFor(logging.DEBUG):
                _LOGGER.debug("list_devices response: %s", await resp.text())

            devices_response: list[dict[str, Any]] = await resp.json()
            return [WxmDevice.unmarshal(d) for d in devices_response]

    async def get_device(self, device_id: str) -> WxmDevice:
        """Get the current status for a WeatherXM device."""
        async with await self.client.get(f"me/devices/{device_id}") as resp:
            await self._raise_if_error(resp)

            if _LOGGER.isEnabledFor(logging.DEBUG):
                _LOGGER.debug(
                    "get_device(%s) response: %s", device_id, await resp.text()
                )

            json_data = await resp.json()
            return WxmDevice.unmarshal(json_data)

    async def get_forecast(
        self,
        device_id: str,
        from_date: datetime.date,
        to_date: datetime.date,
        forecast_type: ForecastType = ForecastType.BOTH,
    ) -> WeatherForecast:
        """Get forecast weather data for a WeatherXM device.

        Forecast data may not be provided if the requested date range is too far
        in the future.
        """
        params = {
            "fromDate": from_date.isoformat(),
            "toDate": to_date.isoformat(),
        }
        if forecast_type != ForecastType.BOTH:
            # Parameter is named "exclude", but appears to actually be an "include" list
            params["exclude"] = "hourly" if ForecastType.HOURLY else "daily"

        async with await self.client.get(
            f"me/devices/{device_id}/forecast", params=params
        ) as resp:
            await self._raise_if_error(resp)

            if _LOGGER.isEnabledFor(logging.DEBUG):
                _LOGGER.debug(
                    "get_forecast(%s) response: %s", params, await resp.text()
                )

            json_data = await resp.json()
            return WeatherForecast.unmarshal(json_data)

    async def get_latest_rewards(self, device_id: str) -> DeviceRewards:
        """Get the latest rewards data for a WeatherXM device."""
        async with await self.client.get(f"devices/{device_id}/rewards") as resp:
            await self._raise_if_error(resp)

            if _LOGGER.isEnabledFor(logging.DEBUG):
                _LOGGER.debug(
                    "get_latest_rewards(%s) response: %s", device_id, await resp.text()
                )

            json_data = await resp.json()
            return DeviceRewards.unmarshal(json_data)

    async def _raise_if_error(self, resp: aiohttp.ClientResponse) -> None:
        """Raise an appropriate exception if an error response was received.

        Does nothing if the respose was OK.
        """
        if resp.ok:
            return

        error_message = f"Unknown response status: {resp.status}"
        if resp.content_type == "application/json":
            json_data = await resp.json()
            error_message = json_data["message"]

        match resp.status:
            case 400:  # Bad request
                # This error code will be raised if the input parameters were invalid
                # so we raise a ValueError for this.
                raise ValueError(error_message)
            case 401:  # Unauthorized request
                raise AuthenticationError(error_message)
            case 500:  # Unexpected error
                raise UnexpectedError(error_message)
            case 522:  # CloudFlare connection timeout response
                raise UnexpectedError("Connection timed out")
            case _:
                # This shouldn't occur
                raise UnexpectedError(error_message)



"""An SDK for querying WeatherXM data."""

from .api import (
    AuthenticationError,
    ForecastType,
    UnexpectedError,
    WxmApi,
    WxmClient,
    WxmError,
)
from .model import (
    BatteryState,
    DailyForecast,
    DeviceRelation,
    DeviceRewards,
    ForecastForDate,
    HourlyForecast,
    HourlyWeatherData,
    Location,
    TokenSummary,
    WeatherForecast,
    WxmDevice,
)

# Explicitly export the public API to satisfy mypy strict type checking.
# fmt:off
__all__ = [
    # API
    "WxmError",
    "AuthenticationError",
    "UnexpectedError",
    "WxmApi",
    "WxmClient",
    "ForecastType",
    # Model
    "BatteryState",
    "DailyForecast",
    "DeviceRelation",
    "DeviceRewards",
    "ForecastForDate",
    "HourlyForecast",
    "HourlyWeatherData",
    "Location",
    "TokenSummary",
    "WeatherForecast",
    "WxmDevice",
]
# fmt:on
